from typing import Optional, Tuple

import discord
from discord import ui

from .component import Component
from .utils import _call_any


class Message:
    def __init__(self,
                 content: Optional[str] = None,
                 embed: Optional[discord.Embed] = None,
                 component: Optional[Component] = None):
        self.content = content
        self.embed = embed
        self.component = component
        self.appear_func = None
        self.disappear_func = None

    async def send(self, channel: discord.abc.Messageable) -> Tuple[ui.View, discord.Message]:
        view = self.component.make_view()
        return view, await channel.send(content=self.content, embed=self.embed, view=view)

    async def update(self, message: discord.Message, other: 'Message') -> Optional[ui.View]:
        kwargs = {}
        if self.content != other.content:
            kwargs['content'] = other.content
            self.content = other.content
        if self.embed != other.embed:
            kwargs['embed'] = other.embed
            self.embed = other.embed
        if self.component != other.component:
            kwargs['view'] = other.component.make_view()
            self.component = other.component
        if not kwargs:
            return None

        await message.edit(**kwargs)
        return kwargs.get('view', None)

    def on_appear(self, func: callable):
        self.appear_func = func
        return self

    def on_disappear(self, func: callable):
        self.disappear_func = func
        return self

    async def appear(self):
        if self.appear_func is not None:
            await _call_any(self.appear_func)

    async def disappear(self):
        if self.disappear_func is not None:
            await _call_any(self.disappear_func)



from typing import Optional, Union

import discord
from discord import ui
from discord import ButtonStyle, PartialEmoji

from .item import Item
from .utils import _call_any


class CustomButton(ui.Button):
    def __init__(
            self,
            *,
            style: ButtonStyle,
            label: str,
            disabled: bool = False,
            custom_id: Optional[str] = None,
            url: Optional[str] = None,
            emoji: Optional[Union[str, PartialEmoji]] = None,
            group: Optional[int] = None,
            callback: Optional[callable] = None
    ):
        super(CustomButton, self).__init__(
            style=style,
            label=label,
            disabled=disabled,
            custom_id=custom_id,
            url=url,
            emoji=emoji,
            group=group
        )
        self.callback_func = callback

    async def callback(self, interaction: discord.Interaction):
        if self.callback_func is None:
            return
        await _call_any(self.callback_func, interaction)


class Button(Item):
    def __init__(
            self,
            label: str,
            style: ButtonStyle = ButtonStyle.primary,
            disabled: bool = False,
            url: Optional[str] = None,
            emoji: Optional[Union[str, PartialEmoji]] = None,
    ):
        self._style = style
        self._label = label
        self._disabled = disabled
        self._url = url
        self._emoji = emoji

        self._group = None

        self.func = None

    def __eq__(self, other: 'Button'):
        return self.to_dict() == other.to_dict()

    def to_discord(self):
        return CustomButton(
            label=self._label,
            style=self._style,
            disabled=self._disabled,
            url=self._url,
            emoji=self._emoji,
            group=self._group,
            callback=self.func
        )

    def to_dict(self):
        return {
            'style': self._style,
            'label': self._label,
            'disabled': self._disabled,
            'emoji': self._emoji,
            'group': self._group,
            'callback': id(self.func)
        }

    def on_click(self, func: callable) -> 'Button':
        self.func = func

        return self

    def style(self, style: discord.ButtonStyle) -> 'Button':
        self._style = style
        return self

    def label(self, label: str) -> 'Button':
        self._label = label
        return self

    def disabled(self, disabled: bool = False) -> 'Button':
        self._disabled = disabled
        return self

    def url(self, url: str) -> 'Button':
        self._url = url
        return self

    def emoji(self, emoji: Union[str, PartialEmoji]) -> 'Button':
        self._emoji = emoji
        return self



import asyncio
from typing import Any

import discord


async def _call_any(func, *args, **kwargs):
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    return func(*args, **kwargs)


def async_interaction_partial(func, *args, **kwargs):
    async def callback(interaction: discord.Interaction) -> Any:
        return await func(interaction, *args, **kwargs)
    return callback


def interaction_partial(func, *args, **kwargs):
    def callback(interaction: discord.Interaction) -> Any:
        return func(interaction, *args, **kwargs)

    return callback



from typing import Optional

from .combine import State, ObservedObject
from .message import Message

from discord import ui
from discord.ext import commands


class View:
    def __init__(self):
        self._watch_variables = []
        self.bot = None
        self.message = None
        self.discord_message = None
        self.view: Optional[ui.View] = None
        self.listeners = []

    def add_listener(self, func, name=None):
        if not isinstance(self.bot, commands.Bot):
            raise ValueError("bot must be commands.Bot")

        self.bot.add_listener(func, name)

    def remove_listener(self, func, name=None):
        if not isinstance(self.bot, commands.Bot):
            raise ValueError("bot must be commands.Bot")

        self.bot.remove_listener(func, name)

    @staticmethod
    def listen(name=None):
        def decorator(func):
            func.__listener__ = True
            func.__listener_name__ = name or func.__name__
            return func
        return decorator

    async def body(self) -> Message:
        return Message()

    async def start(self, bot, channel):
        self.bot = bot
        self.message = await self.body()
        self.view, self.discord_message = await self.message.send(channel)
        await self.message.appear()
        for name in dir(self):
            member = getattr(self, name, None)
            if hasattr(member, "__listener__"):
                self.add_listener(member, getattr(member, "__listener_name__", None))
                self.listeners.append(member)

    async def stop(self):
        for listener in self.listeners:
            self.remove_listener(listener, getattr(listener, "__listener_name__", None))

        await self.message.disappear()
        if self.view is not None:
            self.view.stop()

    async def update(self):
        v = await self.message.update(self.discord_message, (await self.body()))
        if v is not None:
            self.view = v

    def __setattr__(self, key, value):
        if key == "_watch_variables":
            object.__setattr__(self, key, value)
            return

        if isinstance(value, ObservedObject):
            value.view = self

        if not hasattr(self, key):
            if isinstance(value, State):
                self._watch_variables.append(key)
                object.__setattr__(self, key, value.value)
                return
        if isinstance(value, State):
            value = value.value

        if key in self._watch_variables:
            object.__setattr__(self, key, value)
            self.bot.loop.create_task(self.update())
            return

        object.__setattr__(self, key, value)



from typing import Any

from .published import Published


class ObservedObject:
    def __init__(self) -> None:
        self._watch_variables = []
        self.view = None

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "_watch_variables":
            object.__setattr__(self, key, value)
            return

        if not hasattr(self, key):
            if isinstance(value, Published):
                self._watch_variables.append(key)
                object.__setattr__(self, key, value.value)
                return
        if isinstance(value, Published):
            value = value.value

        if key in self._watch_variables:
            object.__setattr__(self, key, value)
            self.view.bot.loop.create_task(self.view.update())
            return

        return object.__setattr__(self, key, value)
